#ifdef __MACOSX_CORE__
  #include <GLUT/glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include "RtAudio.h"
#include "chuck_fft.h"

using namespace std;

typedef struct Cube{
  double x;
  double y;
  double z;
}Cube;


// =======================
// = Function prototypes =
// =======================
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void initialize( );
void drawSphere(double r, int lats, int longs);
void drawSpiral();
void stft(int insize, int fftsize, int hopsize);
void shootingCube(Cube  cube);

// our datetype
#define SAMPLE double
// corresponding format for RtAudio
#define MY_FORMAT RTAUDIO_FLOAT64
// sample rate
#define MY_SRATE 44100
// number of channels
#define MY_CHANNELS 1
// for convenience
#define MY_PIE 3.14159265358979

// initial window size (in pixels)
GLsizei g_width = 1200; 
GLsizei g_height = 1000;

//global buffer for signal
SAMPLE * g_buffer = NULL;
long g_bufferSize;

//old buffers to store signal info
SAMPLE * g_oldbuf = NULL;
SAMPLE * g_oldbuf2 = NULL;
SAMPLE * g_oldbuf3 = NULL;
SAMPLE * g_oldbuf4 = NULL;

//global buffer for spectrum
SAMPLE * g_spectrum = NULL;

//buffers to hold old spectral info
SAMPLE * g_oldspec = NULL;
SAMPLE * g_oldspec2 = NULL;
SAMPLE * g_oldspec3 = NULL;
SAMPLE * g_oldspec4 = NULL;

//set number of cube asteroids
int g_numCubes = 50;

//create space for 5 cubes
Cube * cubes = new Cube[g_numCubes];

//amplitude scalar for waveform
int g_crazy = 1.0;

//amplitude scalar for fft sun
int g_specCrazy = 1.0;

//global variable to move in and out of spiral
float g_enterSpiral = -45.0;

//global variable to automatically enter and exit spiral
int g_autoSpiral = 1;

//variable to hold average amplitude of buffer
double g_average = 0.0;

//variable for threshold of shooting cubes
double g_threshold = 0.5;

//variable for z location of cube
double g_zoom = -20.0;

//-----------------------------------------------------------------------------
// name: callme()
// desc: audio callback
//-----------------------------------------------------------------------------
int callme( void * outputBuffer, void * inputBuffer, unsigned int numFrames,
            double streamTime, RtAudioStreamStatus status, void * data )
{
    // cast!
    SAMPLE * input = (SAMPLE *)inputBuffer;
    
    // fill
    for( int i = 0; i < numFrames; i++ ){

        // assume mono
        g_buffer[i] = input[i];
	g_average += fabs(g_buffer[i]);
    }
    
    //find average of buffer
    g_average = g_average / g_bufferSize;

    //calculate the fft
    stft(g_bufferSize, g_bufferSize, 256);
    
    return 0;
}

//-----------------------------------------------------------------------------
// Entry point
//-----------------------------------------------------------------------------
int main ( int argc, char ** argv ){

  //display the start menu
  cout << "Welcome to Sun Music!" << endl << "Created by Woody Herman" << endl;
  cout << "In order to play music through Sun Music, please connect the output of your JACK compatible music player to the RtApiJACK option that appears in the 'connect' window of QjackCtl" << endl;
  cout << endl << endl << "Below are several keys to play around with while watching Sun Music (must click on window before pressing). Enjoy!" <<endl << endl;
  cout << "'e' - Toggle auto spiraling, initially on." << endl;
  cout << "'i' - Send out spirals of sound from the sun." << endl << "'k' - Retract the spirals." << endl;
  cout << "'u' - Increase ampiltude of spirals." << endl << "'j' - Decrease amplitude of spirals." << endl;
  cout << "'y' - Increase amplitude of the sun." << endl << "'h' - Decrease amplitdude of the sun." << endl;
  cout << "'r' - Increase threshold of asteroids." << endl << "'f' - MOAR ASTEROIDS!" << endl;
  cout << "'q' - Quit." << endl;
  cout << "'z' - Fullscreen." << endl;
  cout << "'x' - Exit fullscreen." << endl;

  // instantiate RtAudio object
  RtAudio audio;
  // variables
  unsigned int bufferBytes = 0;
  // frame size
  unsigned int bufferFrames = 512;
    
  // check for audio devices
  if( audio.getDeviceCount() < 1 ){
    // nopes
    cout << "no audio devices found!" << endl;
    exit( 1 );
  }

  // initialize GLUT
  glutInit( &argc, argv );
  // double buffer, use rgb color, enable depth buffer
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  // initialize the window size
  glutInitWindowSize( g_width, g_height );
  // set the window postion
  glutInitWindowPosition( 50, 50 );
  // create the window
  glutCreateWindow( "MusicVis" );
  //glutEnterGameMode();
  
  // do our own initialization
  initialize();
  

  // let RtAudio print messages to stderr.
  audio.showWarnings( true );
  
  // set input and output parameters
  RtAudio::StreamParameters iParams, oParams;
  iParams.deviceId = audio.getDefaultInputDevice();
  iParams.nChannels = MY_CHANNELS;
  iParams.firstChannel = 0;
  oParams.deviceId = audio.getDefaultOutputDevice();
  oParams.nChannels = MY_CHANNELS;
  oParams.firstChannel = 0;
  
  // create stream options
  RtAudio::StreamOptions options;
  
  // go for it
  try {
    // open a stream
    audio.openStream( &oParams, &iParams, MY_FORMAT, MY_SRATE, &bufferFrames, &callme, (void *)&bufferBytes, &options );
  }
  catch( RtError& e ){
    // error!
    cout << e.getMessage() << endl;
    exit( 1 );
  }

  // compute
  bufferBytes = bufferFrames * MY_CHANNELS * sizeof(SAMPLE);
  
  // allocate global buffers
  g_bufferSize = bufferFrames;
  g_buffer = new SAMPLE[g_bufferSize];
  g_spectrum = new SAMPLE[g_bufferSize];
  g_oldbuf = new SAMPLE[g_bufferSize];
  g_oldbuf2 = new SAMPLE[g_bufferSize];
  g_oldbuf3 = new SAMPLE[g_bufferSize];
  g_oldbuf4 = new SAMPLE[g_bufferSize];
  g_oldspec = new SAMPLE[g_bufferSize];
  g_oldspec2 = new SAMPLE[g_bufferSize];
  g_oldspec3 = new SAMPLE[g_bufferSize];
  g_oldspec4 = new SAMPLE[g_bufferSize];

  //start it up
  try{
    //start audio stream
    audio.startStream();
  
    // let GLUT handle the current thread from here
    glutMainLoop();

    //stop the stream when visual is stopped
    audio.stopStream();
  }
  catch(RtError& e){
    //print out error message
    cout << e.getMessage() << endl;
    goto cleanup;
  }

 cleanup:
  //close stream if open
  if(audio.isStreamOpen())
    audio.closeStream();

  //finito
  return 0;
}

//-----------------------------------------------------------------------------
// Name: initialize( )
// Desc: sets initial OpenGL states
//       also initializes any application data
//-----------------------------------------------------------------------------
void initialize()
{
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
        
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

  
    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );

    // Enable transparency
    glEnable(GL_BLEND);

    //initialize the old buffers with zeros
    for (int i = 0; i < g_bufferSize; i++){
      g_oldbuf[i] = 0;
      g_oldbuf2[i] = 0;
      g_oldbuf3[i] = 0;
      g_oldbuf4[i] = 0;
      g_oldspec[i] = 0;
      g_oldspec2[i] = 0;
      g_oldspec3[i] = 0;
      g_oldspec4[i] = 0;
    }

    //start random seed
    srand (time(NULL));

    //initialize starting location of cubes to zero
    for(int j = 0; j < g_numCubes; j++){
      cubes[j].x = (((rand() * 2.0) /((float) RAND_MAX)) - 1.0) * 1.2;
      cubes[j].y = (((rand() * 2.0) / ((float) RAND_MAX)) - 1.0) * 1.4;
      cubes[j].z = -20.0;
    }

}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    g_width = (GLsizei)w; 
    g_height = (GLsizei)h;
    // map the view port to the client area
    glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
    
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45, (GLfloat)g_width / (GLfloat)g_height, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, 3.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );

}


//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
    case 'Q':
    case 'q':
        exit(1);
        break;
    case 'u':
      g_crazy += 1;
      break;
    case 'j':
      g_crazy -= 1;
      break;
    case 'y':
      g_specCrazy += 1;
      break;
    case 'h':
      g_specCrazy -= 1;
      break;
    case 'i':
      g_enterSpiral += 0.1;
      break;
    case 'k':
      g_enterSpiral -= 0.1;
      break;
    case 'e':
      if(g_autoSpiral == 0)
	g_autoSpiral += 1;
      else
	g_autoSpiral = 0;
      break;
    case 'r':
      g_threshold += 0.1;
      if(g_threshold > 1.0)
	g_threshold = 0.0;
      break;
    case 'f':
      g_threshold -= 0.1;
      if(g_threshold < 0.0)
	g_threshold = 1.0;
      break;
    case 'z':
      glutFullScreen();
      break;
    case 'x':
      glutReshapeWindow(g_width, g_height);
      break;
    }

    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    // animation parameters
    static float cameraangle = 180;
    static float zrot = 0.0;
    static float poop = 0.0;
    static float angle = 0;
    zrot -= 0.1;
    angle -= 0.1;
    if ( angle > 360) angle -= 360;
    poop += 0.001;
    
    //loop through number of cubes
    for(int i = 0; i < g_numCubes; i++){
      //move cube forward
      cubes[i].z += 0.1;
      
      //if cube has reached the end, send it back to the start
      if(cubes[i].z > 2.0){
	cubes[i].z = -20.0;
      }
      
      //if cube is at the start give it a new random location
      if(cubes[i].z == -20.0){
	cubes[i].x = (((rand() * 2.0) /((float) RAND_MAX)) - 1.0) * 1.2;
	cubes[i].y = (((rand() * 2.0) / ((float) RAND_MAX)) - 1.0) * 1.4;
      }
    }
          
    //if auto enter spiral is on, enter and exit the spiral
    if(g_autoSpiral == 1){
      g_enterSpiral = ((cos(poop) + 1) / 2) * -45.0;
    }

    //push the matrix
    glPushMatrix();

      // position the view point
      gluLookAt( 0.0f, 0.0f, g_enterSpiral, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
      glRotatef(cameraangle, 0.0, 1.0, 0.0);
      glRotatef(zrot, 0.0, 0.0, 1.0);
    
      // line width of spiral
      glLineWidth( 5.0 );

      drawSpiral();
    
    // pop
    glPopMatrix();


    // save the current transformation
    glPushMatrix();

        //rotate the sphere
        glRotatef( -angle, 0, 1 , 0);
        glRotatef( -angle, 0, 0 , 1);

        // Start drawing the sphere 
	drawSphere(0.3, 20, 20);
    	
    // restore state
    glPopMatrix();
    
    if(g_average > g_threshold){
      //save current transformation
      glPushMatrix();
      
      for(int j = 0; j < g_numCubes; j++){ 
	shootingCube(cubes[j]);
      }
      //restore matrix
      glPopMatrix();
    }
      
    // flush!
    glFlush( );
      
    // swap the double buffer
    glutSwapBuffers( );
}

//---------------------------------------------------------------------------------------
//void drawSphere
//creates a sphere TO DO: make colors at vertices change in a manner similar to spectogram
//---------------------------------------------------------------------------------------
void drawSphere(double r, int lats, int longs) {
 int i, j;
 for(i = 0; i <= lats; i++) {
   double lat0 = M_PI * (-0.5 + (double) (i - 1) / lats);
   double z0  = sin(lat0);
   double zr0 =  cos(lat0);
   
   double lat1 = M_PI * (-0.5 + (double) i / lats);
   double z1 = sin(lat1);
   double zr1 = cos(lat1);
   
   glBegin(GL_QUAD_STRIP);
   for(j = 0; j <= longs; j++) {
     double lng = 2 * M_PI * (double) (j - 1) / longs;
     double x = cos(lng);
     double y = sin(lng);
     
     glColor4f(1.0, 0.0 , 0.0, 1.0);
     glVertex3f(r * x * zr0, (r * y * zr0) + g_spectrum[j] * g_specCrazy, r * z0);
     
     glColor4f(1.0, 1.0 , 0.0, 1.0);
     glVertex3f((r * x * zr1) + g_spectrum[j] * g_specCrazy,  (r * y * zr1), r * z1);
   }
   glEnd();
 }
}

//--------------------------------------------------------------------------------------
//function drawSpiral: draws a spiral with the waveform on it
//-------------------------------------------------------------------------------------
void drawSpiral(){
    // variables for the x and y positions of the spiral vertices
    GLfloat x,y = 0;
    
    // variable for radius of spiral and icnrement
    GLfloat r = 0;  
    GLfloat dr = 0.002;

    //variable for alpha parameter
    GLfloat alph = 0.0;

    //variable to change color from green to blue
    GLfloat blueness = 0.0;
   
    //integer to hold the number of buffers on spiral
    int numbuffs = 5;

    //blend alphas
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // go
    glBegin( GL_LINE_STRIP );

    // loop through global buffer
      for( int i = 0; i < g_bufferSize; i++ ){
	x = r * cos(r * 10);
	y = r * sin(r * 10);

	//set line color
	glColor4f(0.0, 1.0, blueness / (g_bufferSize*numbuffs), 1 - (alph / (g_bufferSize*numbuffs)));
	glVertex3f(x, y, r+(g_buffer[i] * g_crazy));
	
	// increment x
	r += dr;

	//increment alpha and blueness
	alph += 1;
	blueness += 1;
      }
      for(int i = 0; i  < g_bufferSize; i++){
	x = r * cos(r * 10);
	y = r * sin(r * 10);
	  
	//set line color
	glColor4f(0.0, 1.0, blueness / (g_bufferSize*numbuffs), 1 - (alph / (g_bufferSize*numbuffs)));
	glVertex3f(x, y, r+(g_oldbuf[i] * g_crazy));
	
	// increment x
	r += dr;

	//increment alpha and blueness
	alph += 1;
	blueness += 1;

      }
      for(int i = 0; i  < g_bufferSize; i++){
	x = r * cos(r * 10);
	y = r * sin(r * 10);

	//set line color
	glColor4f(0.0, 1.0, blueness / (g_bufferSize*numbuffs), 1 - (alph / (g_bufferSize*numbuffs)));
	glVertex3f(x, y, r+(g_oldbuf2[i] * g_crazy));
	
	// increment x
	r += dr;

	//increment alpha and blueness
	alph += 1;
	blueness += 1;

      }
      for(int i = 0; i  < g_bufferSize; i++){
	x = r * cos(r * 10);
	y = r * sin(r * 10);

	//set line color
	glColor4f(0.0, 1.0, blueness / (g_bufferSize*numbuffs), 1 - (alph / (g_bufferSize*numbuffs)));
	glVertex3f(x, y, r+(g_oldbuf3[i] * g_crazy));
	
	// increment x
	r += dr;

	//increment alpha and blueness
	alph += 1;
	blueness += 1;
      }
      for(int i = 0; i  < g_bufferSize; i++){
	x = r * cos(r * 10);
	y = r * sin(r * 10);

	//set line color
	glColor4f(0.0, 1.0,  blueness / (g_bufferSize*numbuffs), 1 - (alph / (g_bufferSize*numbuffs)));
	glVertex3f(x, y, r+(g_oldbuf4[i] * g_crazy));
	
	// increment x
	r += dr;

	//increment alpha and blueness
	alph += 1;
	blueness += 1;

	//shuffle info on to next buffers
	g_oldbuf4[i] = g_oldbuf3[i];
	g_oldbuf3[i] = g_oldbuf2[i];
	g_oldbuf2[i] = g_oldbuf[i];
	g_oldbuf[i] = g_buffer[i];
      }
    // done
    glEnd();
}


void stft( int insize, int fftsize, int hopsize){

  //average together several buffers of spectral data for smoothness
  for (int k = 0; k < g_bufferSize; k++){
    g_spectrum[k] = (g_oldspec[k] + g_oldspec2[k] + g_oldspec3[k] + g_oldspec4[k]) / 4;
    
    //move old data into proper buffers
    g_oldspec4[k] = g_oldspec3[k];
    g_oldspec3[k] = g_oldspec2[k];
    g_oldspec2[k] = g_oldspec[k];
  }
  
  //array to hold windowed input
  float * sigframe;
  sigframe = new float[fftsize * 2];

  //int to hold output sample count
  int posout = 0;

  //perform windowing function. here i am using a square window (aka none)
  //this was a styleistic choice. more sidelobes means higher amplitudes 
  //at more frequency bins which means a more bumpy pot-smokin sphere
  for(int i = 0; i < fftsize; i++){
    
    //if we reach the end of the array, pad the remaining portion of the 
    //fft array with zeros
    if(i < insize)
      sigframe[i] = g_buffer[i];
    else
      sigframe[i] = 0;
  }
  
  //perform the fft
  rfft(sigframe, fftsize, 1);
  
  for(int j = 0; j < fftsize; j++, posout++){
    //create output array
    g_oldspec[posout] = sigframe[j];
  }
}

void shootingCube(Cube  cube){

  glTranslatef(cube.x*2, cube.y*2, cube.z);

  //set color of the cube
  glColor4f(0.8, 0.8, 1.0, 0.6);
    
  //draw the cube
  glutSolidCube(0.1);

}
